#ifdef _WIN32
#pragma once
#pragma warning( push )
//4996 = strtok is unsafe, try to use startok_s
#pragma warning(disable: 4996)
#endif

#include "main.h"
#include <set>
#include <iostream>
#include <sstream>

namespace Time{
#include <utils/hr_time.h>
}

using namespace std;

typedef std::set<mpz_class> GraphSet;

int usage()
{
  cout << "Usage: laman n <options> \n this will compute (recursively) the number of laman graphs with\
      1,2,...,n nodes. The number n should be given and should be greater than 3, <options> are optional\n";
  return 1;
}

inline mpz_class convert_to_gmp(const std::vector<bool>& a)
{
  mpz_class out = 0;
  for (size_t i = a.size(); i --> 0; )
    if (a[i]) mpz_setbit(out.get_mpz_t(),i);
  return out;
}

inline vector<bool> convert_to_vecb(size_t n, const mpz_class& a)
{
  vector<bool> out(n,false);
  for (size_t i=0; i<n; ++i)
    if (mpz_tstbit(a.get_mpz_t(),i)) out[i]=true;
  return out;
}

//todo: remove need for vector<bool>
int main(int argc, char *argv[])
{ 
  if (argc==1) return usage();
  int total_nodes = atoi(strtok(argv[1], " \t"));      
  if (total_nodes < 4) return usage();
  
  bool write_it = false;
  bool debug = false;
  if (argc > 2) 
  {
     write_it = (strcmp("w",argv[2])==0);
     debug = (strcmp("d",argv[2])==0);
  }
  if (write_it) cout << "Writing output\n";

  
  cout << "Size of size_t in this computer: " << sizeof(size_t) << " bytes " << endl;
  cout << "Size of mpz_class in this computer: " << sizeof(mpz_class) << " bytes " << endl;
  cout << "Size of sparsegraph in this computer: " << sizeof(sparsegraph) << " bytes " << endl;

  mpz_class gr(7);
  GraphSet orig_set, new_set;  
  cout << "Max size a set can hold in your system: " << orig_set.max_size() << endl;
  new_set.insert(gr);
  std::pair<GraphSet::iterator,bool> insert_info;
  //ofstream ofile;
  FILE *ofile;
  
  Time::CStopWatch csw;
  double time = 0.0;
  
  mpz_class gmptmp = 0;
  for (int n = 3; n< total_nodes; ++n)
  {  
    csw.startTimer();
    orig_set = new_set;
    new_set.clear();
    if (write_it) 
    {
      stringstream sstr;
      sstr << "laman_" << n << ".bin";
      ofile = fopen(sstr.str().c_str(),"wb");
    }

    //uncomment pragma for working
    for (GraphSet::iterator g = orig_set.begin(); g!= orig_set.end(); ++g)
    {
      if (debug && n==5) cout << "(Debug) the graph is " << *g << endl;
      vector<bool> gb = convert_to_vecb(n*(n-1)/2,*g); //number of compact. adj matrix entries must be given.
      //henneberg 1
      //#pragma omp parallel for
      for (int i=0; i<n-1; ++i)
        for (int j=i+1; j<n; ++j)
        {
          vector<bool> tmp = canonical_form(henneberg_1(gb,n,i,j),n+1);
          
          //#pragma omp critical
          gmptmp = convert_to_gmp(tmp);
          insert_info = new_set.insert(gmptmp);
          if (write_it && insert_info.second)  mpz_out_str(ofile,10, gmptmp.get_mpz_t());//mpz_out_raw(ofile, gmptmp.get_mpz_t());
        }

      //henneberg 2
      //#pragma omp parallel for
      for (int e1=0; e1<n-1; ++e1)
        for (int e2=e1+1; e2<n; ++e2)
        {
          if (gb[idx_flat(e2,e1)])
          {
            for (int i=0; i<n-1; ++i)
            {
              if (i==e1 || i==e2) continue;
              std::vector<bool> tmp = canonical_form(henneberg_2(gb,n,e1,e2,i),n+1);

              //#pragma omp critical
              gmptmp = convert_to_gmp(tmp);
              insert_info = new_set.insert(gmptmp);
              if (write_it && insert_info.second)  mpz_out_str(ofile,10, gmptmp.get_mpz_t()); //mpz_out_raw(ofile, gmptmp.get_mpz_t());
            }
          }
        }
    }
    csw.stopTimer();
    time += csw.getElapsedTime();
    cout << "Size of laman graphs with " << n+1 <<" nodes " <<  "after both hennebergs : " << new_set.size() << endl;
    cout << "Time (cummulative) elapsed for this  computation: " << time << " seconds\n";
    if (write_it) fclose(ofile);
  }
    
  return 0;
}

#ifdef _WIN32
#pragma warning(pop)
#endif
